from __future__ import annotations

import copy
import os
import shutil
from typing import Optional, Callable, Any, Dict

from yacs.config import CfgNode as ConfigurationNode
from mlmnemonist.processing_pipeline import Pipeline
from mlmnemonist.runner_cache import RunnerCache


class ExperimentRunner:
    """
    This object type is the basic object you use to run an experiment.
    It is flexible in terms of how the experiment is implemented and it is created for interactive running of experiments.
    You can separate the dataloading process from you training or testing process and you can implement a run function
    using the factory method.

    Each experiment runner has an experiment configuration file associated with it that feeds the information needed
    to conduct an experiment. You can interactively reload the configuration file mid-experiment and re-run everything.
    You can also reload the data using the load_in_dataframe option.

    As an example, checkout the 'RandomForestRuntime.ipynb' in the notebooks section to checkout an experiment runner
    that runs cv_folds on a random forest.

    This object is coupled with a FACTORY and is better to be instantiated using the factory object defined in this
    file. That way, many of the configurations will be preset to the values in the .env.
    """

    def add_config(self,
                   cfg_base: Optional[ConfigurationNode] = None,
                   cfg_path: Optional[str] = None):
        self._has_cfg = False
        self.cfg_path = None
        if cfg_base is not None:
            self._has_cfg = True
            self.cfg = copy.deepcopy(cfg_base)
            if cfg_path is not None:
                self.cfg_path = cfg_path
                self.cfg.merge_from_file(self.cfg_path)

    def __init__(self, experiment_dir: str, checkpoint_dir: str, verbose: int = 0,
                 cache_token: Optional[str] = None, cfg_path: Optional[str] = None,
                 cfg_base: Optional[ConfigurationNode] = None,
                 secret_root: Optional[str] = None,
                 meta_cache: Optional[RunnerCache] = None) -> None:
        self._secret_root = secret_root

        self.add_config(cfg_base, cfg_path)

        self.verbose = verbose

        self.experiment_dir = experiment_dir
        if not os.path.exists(self.experiment_dir):
            raise FileNotFoundError(
                f"File {self.experiment_dir} not found! Maybe you have deleted the experiment ...\n"
                f"Make sure to delete everything in checkpoint directory {checkpoint_dir} too before re-running")

        self.checkpoint_dir = checkpoint_dir

        self._outputs: Dict[str, str] = {}

        self._implemented_run: Optional[Callable[[ExperimentRunner, ...], Any]] = None
        self._META_CACHE = meta_cache
        self.recurring_pipeline = Pipeline(self._META_CACHE,
                                           'RUNNER_RECURRING_PIPELINE_META')
        self.preprocessing_pipeline = Pipeline(self._META_CACHE,
                                               'RUNNER_PREP_PIPELINE_META')

        self.CACHE = RunnerCache(directory=checkpoint_dir, token=cache_token)

    def __str__(self) -> str:
        """
        return a string containing the type of runner
        their cache token,
        configuration (if available)
        the prepcrocessing pipeline
        the recurring pipeline
        and the run function (if implemented)
        """
        ret = f'Runner at {self.experiment_dir}\n'
        ret += f'\t - cache token: {self.CACHE.TOKEN}\n'

        if self._has_cfg:
            ret += f'\t - configurations at: {self.cfg_path}\n'
        else:
            ret += f'\t - no configuration file specified!'

        ret += f'\t - preprocessings functions {self.preprocessing_pipeline}\n'
        ret += f'\t - recurring pipeline {self.recurring_pipeline}\n'
        if self._implemented_run is None:
            ret += f'\t - Run function not implemented!'
        else:
            ret += f'\t - Run function: {self._implemented_run.__name__}\n'
        return ret

    def reveal_true_path(self, path: str) -> str:
        return os.path.join(self._secret_root, path)

    def merge_cfg(self, cfg_path: str) -> None:
        self.cfg_path = cfg_path
        self.reload_cfg()

    def reload_cfg(self) -> None:
        """
        Run this module whenever you have made a change to the cfg but you do not wish
        to redo all the previous steps on the runner. For example, when the data is loaded
        onto an experiment runner and some training configurations have changed, there is no
        need to reload the experiment runner and you can simply change the hyperparameters in
        the yaml file and reload.
        """
        if self._has_cfg and self.cfg_path is not None:
            self.cfg.merge_from_file(self.cfg_path)

    def export_logs(self) -> str:
        ret = os.path.join(self.experiment_dir, 'logs-export')
        shutil.make_archive(ret,
                            'zip',
                            self.CACHE.LOGS_DIR)
        if self.verbose > 0:
            print(f"Files being archived in {ret}.zip ...")
        return ret + '.zip'

    def preprocess(self, keep=True) -> None:
        """
        Run all the functions specified for preprocessing in an orderly fasion
        """
        if self.preprocessing_pipeline.function_count == 0 and self.verbose > 0:
            print("No functions in the preprocessing pipeline!")
        self.reload_cfg()
        self.preprocessing_pipeline.run(keep=keep, verbose=self.verbose, runner=self)

    def implement_run(self, run: Callable[[ExperimentRunner, ...], Any]) -> None:
        """
        Use this function to
        """
        self._implemented_run = run
        self._META_CACHE.SET('RUNNER_RUN_META', self._implemented_run)
        self._META_CACHE.SAVE()

    def ADD_OUTPUT(self, file_dir: str, description: str):
        self._outputs[file_dir] = description

    def run(self, *args, **kwargs):
        """
        This function runs an arbitrary method that is specified in the
        factory function. After running it, all the values of the config file plus a timer of the whole runtime
        will be stored in a .yaml file in the experiments directory. This will give us the possibility to re-run
        our experiments by simply specifying the output yaml file as input again.
        """
        if self._implemented_run is None:
            raise NotImplementedError("The run function is not implemented yet!")

        if self.verbose > 0 and self.recurring_pipeline.function_count == 0:
            print("No functions in the recurring pipeline ...")
        self.reload_cfg()
        self.recurring_pipeline.run(keep=True, verbose=self.verbose, runner=self)

        self.CACHE.LOAD()
        ret = self._implemented_run(self, *args, **kwargs)

        # Save files
        if self.verbose > 0:
            print("[DONE] running over!")
            print("\t - saving files ...")

        with open(os.path.join(self.experiment_dir, 'readme.txt'), 'w') as f:
            all_lines = ['This file contains a description on the files available in the experiment\n']
            # Save configurations
            if self._has_cfg and self.cfg is not None and self.cfg_path is not None:
                name = '.'.join(os.path.basename(self.cfg_path).split('.')[:-1])
                self.cfg.dump(
                    stream=open(os.path.join(self.experiment_dir, f'{name}-output.yaml'), 'w'))
                shutil.copyfile(self.cfg_path, os.path.join(self.experiment_dir, f'{name}-input.yaml'))
                all_lines.append(f"\t{name}-output.yaml : Contains the configurations after ending the runner.\n")
                all_lines.append(f"\t{name}-input.yaml : Contains the configurations when starting the runner.\n"
                                 f"\t You can feed the same file again as configurations and gain the same results.\n")
            # Save file descriptions in readme.txt
            all_lines.append("Output files and their descriptions:\n")
            all_lines += [f'\t{x} : {self._outputs[x]}\n' for x in self._outputs.keys()]
            f.writelines(all_lines)
        return ret



import os
import shutil
import warnings

import torch
import torch.nn as nn
from typing import Optional, Union, Dict, List, Any
import pickle
from yacs.config import CfgNode as ConfigurationNode
import yacs
import yacs.config

MAX_CACHE_SIZE: int = 20


def _get_all_tokens(directory: str) -> List[str]:
    all_names = set()
    for f in os.listdir(directory):
        if '-MLM-CACHE-TOK' not in f:
            continue
        name = f.split('-MLM-CACHE-TOK')[0] + '-MLM-CACHE-TOK'
        all_names.add(name)
    return list(all_names)


def _get_new_token(directory: str) -> str:
    all_tokens = _get_all_tokens(directory)
    mex = 0
    while f'{mex}' in all_tokens:
        mex += 1
    return f'{mex}'


class RunnerCache:

    def __init__(self, directory: str, token: Optional[str] = None):
        all_tokens = _get_all_tokens(directory)

        if token not in all_tokens and len(all_tokens) >= MAX_CACHE_SIZE:
            raise Exception("Maximum cache limit reached!\n"
                            f"Remove files from {directory}")
        self._cache_token = token if token is not None else _get_new_token(directory)
        self._cache_token += '-MLM-CACHE-TOK'

        self._cached_primitives: Dict[str, Any] = {}
        self._cached_models: Dict[str, nn.Module] = {}
        self._cached_configs: Dict[str, ConfigurationNode] = {}

        self.directory = directory
        # Create the logfile
        log_dir = os.path.join(self.directory, f'{self._cache_token}-logs')
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

    @property
    def LOGS_DIR(self) -> str:
        return os.path.join(self.directory, f'{self._cache_token}-logs')

    @property
    def TOKEN(self) -> str:
        return self._cache_token.split('-MLM-CACHE-TOK')[0]

    def RESET(self, prompt=True) -> None:
        """
        This function removes all the cached data from disk.
        By defeault, it will prompt the user to ask and make sure they agree to this.
        However, one can use it by setting 'prompt = False' to avoid prompting.
        """
        while prompt:
            resp = input("This will remove all of the variables from your runner's cache ..."
                         "\nAre you sure you want to proceed? [y/n] ")
            if resp == 'y':
                break
            elif resp == 'n':
                return
            else:
                continue

        # Remove everything from the checkpoint
        # directory that starts with the runner prefix
        for f in os.listdir(self.directory):
            real_dir = os.path.join(self.directory, f)
            if f.startswith(self._cache_token):
                if os.path.isfile(real_dir):
                    os.remove(os.path.join(self.directory, f))
                elif f == f'{self._cache_token}-logs':
                    shutil.rmtree(real_dir)

        # Clear up the cached_primitives
        self._cached_primitives = {}
        self._cached_models = {}

    def SAVE(self) -> None:
        """
        Save the whole cache into the disk.
        """
        try:
            # Save the primitives
            with open(os.path.join(self.directory, f'{self._cache_token}-runner_checkpoint_primitives.pkl'), 'wb') as f:
                pickle.dump(self._cached_primitives, f)

            # Save all of the current models
            for name in self._cached_models.keys():
                torch.save(self._cached_models[name].state_dict(),
                           os.path.join(self.directory, f'{self._cache_token}-model.{name}.pth'))
            # Save all the configurations
            for name in self._cached_configs.keys():
                cfg = self._cached_configs[name]
                cfg.dump(stream=open(os.path.join(self.directory, f'{self._cache_token}-cfg.{name}.yaml'), 'w'))

        except KeyboardInterrupt as e:
            warnings.warn("No keyboard interrupt allowed in between cache saving ...")

    def SET_IFN(self, name: str, value: Any) -> Any:
        if name not in self._cached_primitives:
            self._cached_primitives[name] = value
        return self._cached_primitives[name]

    def SET(self, name: str, value: Any) -> Any:
        self._cached_primitives[name] = value
        return self._cached_primitives[name]

    def GET(self, name: str) -> Any:
        if name not in self._cached_primitives:
            raise ModuleNotFoundError(f"primitive with name {name} not found!")
        return self._cached_primitives[name]

    def SET_CFG(self, name: str, value: ConfigurationNode):
        self._cached_configs[name] = value
        return self._cached_configs

    def SET_IFN_CFG(self, name: str, value: Optional[ConfigurationNode]) -> Optional[ConfigurationNode]:
        if name not in self._cached_configs:
            self._cached_configs[name] = value
        return self._cached_configs[name]

    def GET_CFG(self, name: str):
        if name not in self._cached_configs:
            raise Exception(f"Config with name {name} not found!")
        return self._cached_configs[name]

    def SET_IFN_M(self, name: str, model: Optional[nn.Module]) -> Optional[nn.Module]:
        if '.' in name:
            raise NameError(f"Name should not contain dots {name}")
        if name not in self._cached_models:
            self._cached_models[name] = model
        return self._cached_models[name]

    def SET_M(self, name: str, model: nn.Module):
        if '.' in name:
            raise NameError(f"Name should not contain dots {name}")
        self._cached_models[name] = model
        return self._cached_models[name]

    def GET_M(self, name: str):
        if name not in self._cached_models:
            raise ModuleNotFoundError(f"model {name} not found!")
        return self._cached_models[name]

    def LOAD(self):
        for dir in os.listdir(self.directory):
            real_dir = os.path.join(self.directory, dir)
            if dir == f'{self._cache_token}-runner_checkpoint_primitives.pkl':
                # Handle loading the primitives
                with open(real_dir, 'rb') as f:
                    self._cached_primitives = pickle.load(f)
            elif dir.startswith(f'{self._cache_token}-model'):
                # Handle loading models
                model_name = real_dir.split('.')[-2]
                if model_name not in self._cached_models:
                    print(f"Model with name {model_name} has not been initialized in preprocess!")
                self._cached_models[model_name].load_state_dict(torch.load(real_dir))
            elif dir.startswith(f'{self._cache_token}-cfg'):
                cfg_name = real_dir.split('.')[-2]
                self._cached_configs[cfg_name] = yacs.config.load_cfg(open(real_dir, 'r'))



import copy
import os
import sys

from dotenv import load_dotenv, find_dotenv
from yacs.config import CfgNode as ConfigurationNode
import re


def _is_branch_type(cfg: ConfigurationNode):
    """
    :param cfg:
    A ConfigurationNode
    :return:
    returns true if the children are of type MLM_BRANCH_i
    returns false otherwise
    """
    branch_count = 0
    for k in cfg.keys():
        if re.match(r'MLM_BRANCH_\d+', k):
            branch_count += 1

    if branch_count != 0 and branch_count != len(cfg.keys()):
        raise Exception(f"MLM branches incorrect at:\n{cfg}")
    return branch_count > 0


def _merge_cfg(cfg1: ConfigurationNode, cfg2: ConfigurationNode, key_list=None) -> ConfigurationNode:
    if key_list is None:
        key_list = []

    if not isinstance(cfg1, ConfigurationNode):
        return cfg2
    if not isinstance(cfg2, ConfigurationNode):
        return cfg1

    if _is_branch_type(cfg2):
        leaf = False
        if not isinstance(cfg1, ConfigurationNode):
            leaf = True
        base_cfg = copy.deepcopy(cfg1)
        cfg1 = ConfigurationNode()
        for k in cfg2.keys():
            if leaf:
                cfg1[k] = copy.deepcopy(cfg2[k])
            else:
                new_cfg = copy.deepcopy(base_cfg)
                new_cfg = _merge_cfg(new_cfg, cfg2[k], key_list)
                cfg1[k] = new_cfg
    else:
        for k in cfg2.keys():
            if not isinstance(cfg1, ConfigurationNode) or k not in cfg1:
                raise KeyError("Non-existent config key: {}".format(key_list + [k]))
            if not isinstance(cfg2[k], ConfigurationNode):
                cfg1[k] = copy.deepcopy(cfg2[k])
            else:
                cfg1[k] = _merge_cfg(cfg1[k], cfg2[k], key_list + [k])
    return cfg1


def _expand_cfg(cfg: ConfigurationNode, key_list=None):
    if key_list is None:
        key_list = []

    if not isinstance(cfg, ConfigurationNode):
        return [[copy.deepcopy(cfg), []]]
    elif not _is_branch_type(cfg):
        old_variants = []
        for k in cfg.keys():
            new_variants = _expand_cfg(cfg[k], key_list + [k])
            if len(old_variants) == 0:
                for cfg_, code in new_variants:
                    new_cfg = ConfigurationNode()
                    new_cfg[k] = copy.deepcopy(cfg_)
                    old_variants.append([new_cfg, code])
            else:
                accumulate = []
                for cfg_new, code_new in new_variants:
                    for cfg_old, code_old in old_variants:
                        cfg_combined = copy.deepcopy(cfg_old)
                        cfg_combined[k] = copy.deepcopy(cfg_new)
                        accumulate.append([cfg_combined, copy.deepcopy(code_old) + copy.deepcopy(code_new)])
                old_variants = accumulate
        return old_variants
    else:
        variants = []
        for i, k in enumerate(cfg.keys()):
            if isinstance(cfg[k], ConfigurationNode):
                new_variants = _expand_cfg(cfg[k], key_list+[k])
                for var in new_variants:
                    var[1] += [i]
                variants += new_variants
            else:
                variants.append([copy.deepcopy(cfg[k]), [i]])
        return variants


def expand_cfg(cfg_base, cfg_dir: str, save_directory: str):

    load_dotenv(find_dotenv(), verbose=True)  # Load .env
    cfg_dir = os.path.join(os.getenv('MLM_CONFIG_DIR'), cfg_dir)
    save_directory = os.path.join(os.getenv('MLM_EXPERIMENT_DIR'), save_directory)

    with open(cfg_dir, "r") as f:
        other_cfg = cfg_base.load_cfg(f)
    grid_cfg = copy.deepcopy(cfg_base)
    grid_cfg = _merge_cfg(grid_cfg, other_cfg)

    variants = _expand_cfg(grid_cfg)
    if not os.path.exists(save_directory):
        sv_dir = os.getcwd()
        parent_dir, dir_name = os.path.split(save_directory)
        parent_dir = os.path.join(parent_dir)
        os.chdir(parent_dir)
        os.mkdir(dir_name)
        os.chdir(sv_dir)

    mx_code = 1
    for _, code in variants:
        mx_code = max(mx_code, len(code))

    ret = {}
    for cfg, code in variants:
        name = '-'.join([str(c) for c in code])
        for _ in range(mx_code - len(code)):
            name += '-0'
        pth = os.path.join(save_directory, f'{name}-MLM-{os.path.basename(cfg_dir)}')
        ret[name] = pth
        cfg.dump(stream=open(pth, 'w'))
    return ret



import torch.nn as nn

from testing.modeling.cell_embedding import CellEmbedder, CELL_EMBEDDER_REGISTRY
from testing.modeling.compound_embedding import CompoundEmbedder, COMPOUND_EMBEDDER_REGISTRY
from testing.modeling.fusion import DrugConcentrationFusion, FUSION_REGISTRY
from testing.modeling.integration import DrugCellIntegrationModel, INTEGRATION_REGISTRY
from testing.modeling.registry import Registry
from yacs.config import CfgNode as ConfigurationNode

DEEP_DDR_REGISTRY = Registry()


class DrugDoseResponseModel(nn.Module):
    def __init__(self,
                 compound_embedder: CompoundEmbedder,
                 cell_embedder: CellEmbedder,
                 drug_concentration_fusion: DrugConcentrationFusion,
                 drug_cell_integrator: DrugCellIntegrationModel,
                 dropout: float
                 ):
        super(DrugDoseResponseModel, self).__init__()
        if cell_embedder.out_features is None:
            raise AttributeError("cell embedder out feature dimension not specified!")

        if compound_embedder.in_features is None:
            raise AttributeError("molecule embedder out feature dumension no specified!")

        # TODO: add more exceptions for dimensions

        self.compound_embedder = compound_embedder
        self.cell_embedder = cell_embedder
        self.drug_cell_integrator = drug_cell_integrator
        self.drug_concentration_fusion = drug_concentration_fusion

        self.dropout_rate = dropout
        self.linear1 = nn.Linear(self.drug_cell_integrator.out_features, 1000, bias=False)
        self.batchnorm1 = nn.BatchNorm1d(1000)
        self.linear2 = nn.Linear(1000, 500, bias=False)
        self.batchnorm2 = nn.BatchNorm1d(500)
        self.linear3 = nn.Linear(500, 1)


class NoPoolingDDRM(DrugDoseResponseModel):
    def __init__(self, *args, **kwargs):
        super(NoPoolingDDRM, self).__init__(*args, **kwargs)


def prepare_models(cfg_hyper_parameters: ConfigurationNode):
    cell_embedder_type = cfg_hyper_parameters.CELL_EMBEDDER.TYPE
    cell_embedder_cfg = cfg_hyper_parameters.CELL_EMBEDDER.HYPER_PARAMETERS
    cell_embedder = CELL_EMBEDDER_REGISTRY[cell_embedder_type](cell_embedder_cfg)

    drug_embedder_type = cfg_hyper_parameters.DRUG_EMBEDDER.TYPE
    drug_embedder_cfg = cfg_hyper_parameters.DRUG_EMBEDDER.HYPER_PARAMETERS
    drug_embedder = COMPOUND_EMBEDDER_REGISTRY[drug_embedder_type](drug_embedder_cfg)

    fusion_type = cfg_hyper_parameters.FUSION.TYPE
    fusion_cfg = cfg_hyper_parameters.FUSION.HYPER_PARAMETERS
    fusion_model = FUSION_REGISTRY[fusion_type](fusion_cfg, drug_embedder.out_features)

    integration_type = cfg_hyper_parameters.INTEGRATION.TYPE
    integration_cfg = cfg_hyper_parameters.INTEGRATION.HYPER_PARAMETERS
    integration_model = INTEGRATION_REGISTRY[integration_type](integration_cfg, fusion_model.out_features,
                                                               cell_embedder.out_features)
    return cell_embedder, drug_embedder, fusion_model, integration_model


@DEEP_DDR_REGISTRY.register('deep-ddr-v1')
def build_deep_ddr(cfg_hyper_parameters: ConfigurationNode):
    cell_embedder, drug_embedder, fusion_model, integration_model = prepare_models(cfg_hyper_parameters)

    return DrugDoseResponseModel(compound_embedder=drug_embedder,
                                 cell_embedder=cell_embedder,
                                 drug_concentration_fusion=fusion_model,
                                 drug_cell_integrator=integration_model,
                                 dropout=cfg_hyper_parameters.DROPOUT)


@DEEP_DDR_REGISTRY.register('deep-ddr-v2')
def build_deep_ddr_large_vertex_set(cfg_hyper_parameters: ConfigurationNode):
    cell_embedder, drug_embedder, fusion_model, integration_model = prepare_models(cfg_hyper_parameters)

    return NoPoolingDDRM(
        compound_embedder=drug_embedder,
        cell_embedder=cell_embedder,
        drug_concentration_fusion=fusion_model,
        drug_cell_integrator=integration_model,
        dropout=cfg_hyper_parameters.DROPOUT
    )



import random
import time

from config.config_drug import get_cfg_defaults
from mlmnemonist import RunnerFactory, ExperimentRunner
from mlmnemonist.validation_tools import grid_search, grid_search_from_palette
from testing.modeling.drug_combo import DEEP_DDR_REGISTRY


def decoy_prep(runner: ExperimentRunner):
    if runner.verbose > 0:
        print("Running preprocess!")


def decoy_recurring(runner: ExperimentRunner):
    if runner.verbose > 0:
        print("Running recurring!")


def decoy_run(runner: ExperimentRunner):
    iteration_i = runner.CACHE.SET_IFN('iteration_i', 0)
    while iteration_i < 10:
        if runner.verbose > 0:
            print("running runner:", iteration_i)
        time.sleep(0.5)
        iteration_i += 1
        runner.CACHE.SET('iteration_i', iteration_i)
        runner.CACHE.SAVE()
    return random.randint(0, 10)


def config_based_run(runner: ExperimentRunner):
    model = DEEP_DDR_REGISTRY[runner.cfg.DEEP_DDR.MODEL.TYPE](runner.cfg.DEEP_DDR.MODEL.HYPER_PARAMETERS)
    return random.randint(0, 10)


if __name__ == '__main__':
    factory = RunnerFactory()
    try:
        runner = factory.retrieve('unittest2')
    except FileNotFoundError:
        runner = factory.create(experiment_name='build-a-large',
                                verbose=1,
                                cache_token='unittest2')

    runner.preprocessing_pipeline.update_function(decoy_prep)
    runner.preprocess()
    runner.recurring_pipeline.update_function(decoy_recurring)
    # runner.implement_run(decoy_run)
    # runner.run()
    runner.implement_run(config_based_run)
    grid_search_from_palette(runner,
                             cache_token='chiz',
                             verbose=1,
                             save_directory='config/all-conf-drug-tuning',
                             cfg_base=get_cfg_defaults(),
                             cfg_palette_dir='conf-drug-palette.yaml')



# Copyright 2015 Google Inc.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains Django URL patterns used for OAuth2 flow."""

from django.conf import urls

from oauth2client.contrib.django_util import views

urlpatterns = [
    urls.url(r'oauth2callback/', views.oauth2_callback, name="callback"),
    urls.url(r'oauth2authorize/', views.oauth2_authorize, name="authorize")
]

urls = (urlpatterns, "google_oauth", "google_oauth")




from yacs.config import CfgNode as ConfigurationNode

# YACS overwrite these settings using YAML
__C = ConfigurationNode()

# Dummy dataset variables only for testing
__C.DATASET = ConfigurationNode()
__C.DATASET.TRAIN_NAME = 'not-specified!'
__C.DATASET.TEST_NAME = 'not-specified!'

__C.SOLVER = ConfigurationNode()
__C.SOLVER.LR = 0.0
__C.SOLVER.OPTIMIZER_TYPE = 'not-specified!'
__C.SOLVER.DEVICE = 'not-specified!'
__C.SOLVER.METHOD = 'not-specified!'

__C.MODEL = ConfigurationNode()
__C.MODEL.HYPER_PARAMETERS = ConfigurationNode()
__C.MODEL.HYPER_PARAMETERS.IN_FEATURES = -1
__C.MODEL.HYPER_PARAMETERS.H1 = -1
__C.MODEL.HYPER_PARAMETERS.H2 = -1


def get_cfg_defaults() -> ConfigurationNode:
    """
    Get a yacs CfgNode object with default values
    """
    # Return a clone so that the defaults will not be altered
    # It will be subsequently overwritten with local YAML.
    return __C.clone()



import json
import os
from typing import List, Optional, Dict, Any

from mlmnemonist.runner_cache import RunnerCache
from yacs.config import CfgNode as ConfigurationNode

from mlmnemonist import ExperimentRunner
from mlmnemonist.validation_tools import expand_cfg
from mlmnemonist.validation_tools.grid_search import _run_grid_search
from mlmnemonist.validation_tools.utils import get_all_codes


class HyperExperimentRunner:
    def __init__(self,
                 cfg_palette_dir: str,
                 cfg_base: ConfigurationNode,
                 hyper_experiment_path: str,
                 experiment_runners: List[ExperimentRunner],
                 checkpoint_dir: str,
                 cache_token: Optional[str] = None,
                 verbose: Optional[int] = 0,
                 secret_root: Optional[str] = None
                 ):
        self.cfg_base = cfg_base
        self._secret_root = secret_root
        self.hyper_experiment_path = hyper_experiment_path
        if not os.path.exists(self.hyper_experiment_path):
            raise FileNotFoundError(
                f"File {self.hyper_experiment_path} not found! Maybe you have deleted the experiment ...\n"
                f"Make sure to delete everything in checkpoint directory {checkpoint_dir} too before re-running")
        self.all_cfgs_dir = os.path.join(self.hyper_experiment_path, 'all-cfgs')
        if not os.path.exists(self.all_cfgs_dir):
            os.mkdir(self.all_cfgs_dir)
        expand_cfg(cfg_base, cfg_palette_dir, self.all_cfgs_dir)

        self.runners = experiment_runners
        self.verbose = verbose
        self.checkpoint_dir = checkpoint_dir

        self.CACHE = RunnerCache(self.checkpoint_dir, f'{cache_token}')

    def reveal_true_path(self, path: str) -> str:
        return os.path.join(self._secret_root, path)

    def full_search(self, with_preprocess=False, *args, **kwargs) -> Dict[str, Any]:
        # Setup remaining codes
        self.CACHE.LOAD()
        all_codes = self.CACHE.SET_IFN('all_codes', None)
        iteration_i = self.CACHE.SET_IFN('iteration_i', 0)
        codes_dict = self.CACHE.SET_IFN('codes_dict', None)
        score_dict = self.CACHE.SET_IFN('score_dict', {})
        if all_codes is None:
            codes_dict = get_all_codes(self.all_cfgs_dir)
            all_codes = list(codes_dict.keys())
        self.CACHE.SET('all_codes', all_codes)
        self.CACHE.SET('iteration_i', iteration_i)
        self.CACHE.SET('codes_dict', codes_dict)
        self.CACHE.SET('score_dict', score_dict)
        self.CACHE.SAVE()

        # Iterate over all the remaining codes
        while iteration_i < len(all_codes):
            current_code = all_codes[iteration_i]
            self.runners[0].merge_cfg(codes_dict[current_code])
            self.runners[0].verbose = max(0, self.verbose - 1)
            if self.verbose > 0:
                print(f"Iteration no. [{iteration_i + 1}/{len(all_codes)}] "
                      f"-- Running {os.path.split(self.runners[0].cfg_path)[-1]}")
            if with_preprocess:
                self.runners[0].preprocess()
            score = self.runners[0].run(*args, **kwargs)
            score_dict[codes_dict[current_code]] = score
            self.runners[0].CACHE.RESET(prompt=False)
            if self.verbose > 0:
                print("Final score:", score)

            with open(os.path.join(self.all_cfgs_dir, 'seen_scores.json'), 'w') as f:
                json.dump(score_dict, f, indent=4)

            iteration_i += 1

            self.CACHE.SET('all_codes', all_codes)
            self.CACHE.SET('iteration_i', iteration_i)
            self.CACHE.SET('codes_dict', codes_dict)
            self.CACHE.SET('score_dict', score_dict)
            self.CACHE.SAVE()

        return score_dict

    def grid_search(self, with_preprocess: bool = False, *args, **kwargs):
        return _run_grid_search(self.runners[0],
                                self.CACHE,
                                self.verbose,
                                self.cfg_base,
                                self.all_cfgs_dir,
                                with_preprocess, *args, **kwargs)



from mlmnemonist.validation_tools.cfg_utils import expand_cfg
from mlmnemonist.validation_tools.grid_search import grid_search_from_palette, grid_search
from mlmnemonist.validation_tools.grid_search import GRID_SEARCH_CACHING



GOOGLE_APP_AUDIO = "application/vnd.google-apps.audio"
GOOGLE_DOC = "application/vnd.google-apps.document"
GOOGLE_DRAWING = "application/vnd.google-apps.drawing"

GOOGLE_DRIVE_FILE = "application/vnd.google-apps.file"
GOOGLE_DRIVE_FOLDER = "application/vnd.google-apps.folder"

GOOGLE_FORMS = "application/vnd.google-apps.format"
GOOGLE_FUSION_TABLES = "application/vnd.google-apps.fusiontable"
GOOGLE_MY_MAPS = "application/vnd.google-apps.map"
GOOGLE_SLIDES = "application/vnd.google-apps.presentation"
GOOGLE_APPS_SCRIPTS = "application/vnd.google-apps.Scripts"
GOOGLE_SITES = "application/vnd.google-apps.sites"
GOOGLE_SHEETS = "application/vnd.google-apps.spreadsheet"

TEXT = "text/plain"
CSV = "text/csv"
JSON = "application/json"

XLS = "application/vnd.ms-excel"
XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
DOC = "application/msword"
DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
PPT = "application/vnd.ms-powerpoint"
PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"



from abc import ABC
from typing import List
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GATConv, GCNConv
from torch_geometric.nn import global_max_pool

from yacs.config import CfgNode as ConfigurationNode

from testing.modeling.registry import Registry

COMPOUND_EMBEDDER_REGISTRY = Registry()


class CompoundEmbedder(nn.Module):
    def __init__(self, in_features: int):
        super(CompoundEmbedder, self).__init__()
        self.in_features = in_features
        self.out_features = None


class GNNEmbedder(CompoundEmbedder, ABC):
    def __init__(self, in_features: int,
                 activation: str):
        super(GNNEmbedder, self).__init__(in_features)

        assert activation in ['relu', 'elu'], f'Activation {activation} not defined'

        if activation == 'relu':
            self.activation = F.relu
        elif activation == 'elu':
            self.activation = F.elu

        self.layer_count: int = 0

    def forward(self, x, edge_index, batch):
        for i in range(self.layer_count):
            L = self._modules[f'gnn_{i + 1}']
            x = L(x, edge_index)
            x = self.activation(x)
        return x, batch


class GCNEmbedder(GNNEmbedder):
    def __init__(self,
                 in_features: int,
                 d_layers: List[int],
                 activation: str):
        super(GCNEmbedder, self).__init__(in_features, activation)

        # graph drug layers
        prv = in_features
        for i, d in enumerate(d_layers):
            self.add_module(f'gnn_{i + 1}', GCNConv(prv, d))
            prv = d
        self.layer_count = len(d_layers)
        self.out_features = prv


class GATEmbedder(GNNEmbedder):
    def __init__(self,
                 in_features: int,
                 nheads: int,
                 out_features: int,
                 activation: str,
                 dropout: float):
        super(GATEmbedder, self).__init__(in_features, activation)

        self.add_module(f'gnn_1', GATConv(in_features, out_features, heads=nheads, dropout=dropout))
        self.add_module(f'gnn_2', GATConv(out_features * nheads, out_features, dropout=dropout))
        self.out_features = out_features
        self.layer_count = 2


class GlobalPoolingEmbedder(nn.Module):
    def __init__(self, out_features, compound_embedder: GNNEmbedder):
        super(GlobalPoolingEmbedder, self).__init__()
        h = compound_embedder.out_features
        self.fc1 = nn.Linear(h, out_features // 4)
        self.fc2 = nn.Linear(out_features // 4, out_features // 2)
        self.compound_embedder = compound_embedder

    def forward(self, x, edge_index, batch):
        x, batch = self.compound_embedder(x, edge_index, batch)
        x = global_max_pool(x, batch)
        x = self.fc1(x)
        x = F.relu(x)
        x = F.dropout(x, p=self.dropout_rate, training=self.training)

        x = self.fc2(x)
        x = F.relu(x)
        x = F.dropout(x, p=self.dropout_rate, training=self.training)

        return x


def build_gnn_embedder(cfg_hyper_parameters: ConfigurationNode):
    if cfg_hyper_parameters.GNN_TYPE == 'GAT':
        return GATEmbedder(in_features=cfg_hyper_parameters.IN_FEATURES,
                           nheads=cfg_hyper_parameters.N_HEADS,
                           out_features=cfg_hyper_parameters.OUT_FEATURES,
                           activation=cfg_hyper_parameters.ACTIVATION,
                           dropout=cfg_hyper_parameters.DROPOUT)
    elif cfg_hyper_parameters.GNN_TYPE == 'GCN':
        return GCNEmbedder(in_features=cfg_hyper_parameters.IN_FEATURES,
                           d_layers=cfg_hyper_parameters.D_LAYERS,
                           activation=cfg_hyper_parameters.ACTIVATION)
    else:
        raise NotImplementedError(f"GNN type {cfg_hyper_parameters.GNN_TYPE} not implemented!")


@COMPOUND_EMBEDDER_REGISTRY.register('gnn-global-pooling')
def build_gnn_global_pooling_embedder(cfg_hyper_parameters: ConfigurationNode):
    compound_embedder = build_gnn_embedder(cfg_hyper_parameters)
    return GlobalPoolingEmbedder(out_features=cfg_hyper_parameters.POOLING_OUT_FEATURES,
                                 compound_embedder=compound_embedder)


@COMPOUND_EMBEDDER_REGISTRY.register('gnn-no-global-pooling')
def build_gnn_with_no_pooling_embedder(cfg_hyper_parameters: ConfigurationNode):
    return build_gnn_embedder(cfg_hyper_parameters)



import copy
import os
import pickle
import shutil
import warnings
from typing import Optional, Callable, List
from datetime import date
from dotenv import load_dotenv, find_dotenv
from mlmnemonist.runner_cache import RunnerCache

from mlmnemonist.experiment_runner import ExperimentRunner
from yacs.config import CfgNode as ConfigurationNode

from mlmnemonist.validation_tools.hyper_experiment_runner import HyperExperimentRunner


def _get_new_experiment_path(root_experiments_path, experiment_name: Optional[str]):
    """
    Get the full directory of the experiment, if experiment_name is not specified
    then come up with the currently unavailable name!
    """
    # setup experiment name
    middle_name = 'exp' if experiment_name is None else experiment_name

    mex = 0
    while os.path.exists(os.path.join(root_experiments_path, f'{date.today()}-{middle_name}-{mex}')):
        mex += 1
    exp_name = f'{date.today()}-{middle_name}-{mex}'

    experiment_path = os.path.join(root_experiments_path, exp_name)
    if not os.path.exists(experiment_path):
        os.mkdir(experiment_path)
    return experiment_path


class RunnerFactory:
    """
    An experiment runner factory that creates experiment runners and customizes them.

    Use the 'create' method first to create a generic runner and then add on new features to the runner.

    - load_combo_dataset: loads a dataset according to the config file given to it; this directly calls
                            ExperimentRunner.load_in_dataframe
    - load_cv_fold: loads the cv_folds according to the paths given to it in the environment. This function
                        directly calls ExperimentRunner.load_cv_fold
    - implement_run: This function
    """

    _instance_count: int = 0

    def __init__(self,
                 experiment_dir: Optional[str] = None,
                 checkpoint_dir: Optional[str] = None,
                 config_dir: Optional[str] = None,
                 secret_dir: Optional[str] = None,
                 override_singleton: bool = False):
        """
        Loads data from .env and fills up the following:
        - MLM_EXPERIMENT_DIR=The directory containing the experiments
        - MLM_CHECKPOINT_DIR=Directory containing all the checkpoints
        - MLM_CONFIG_DIR=directory containing all the .yaml files
        - MLM_SECRET_ROOT_DIR=secret prefix used for secret paths
        """
        if RunnerFactory._instance_count > 0 and not override_singleton:
            raise Exception("Singleton class RunnerFactory cannot be instantiated more than once!")
        RunnerFactory._instance_count += 1

        load_dotenv(find_dotenv(), verbose=True)  # Load .env

        all_args = [experiment_dir, 'MLM_EXPERIMENT_DIR',
                    checkpoint_dir, 'MLM_CHECKPOINT_DIR',
                    config_dir, 'MLM_CONFIG_DIR',
                    secret_dir, 'MLM_SECRET_ROOT_DIR']
        for i in range(0, len(all_args), 2):
            if all_args[i] is None:
                all_args[i] = os.getenv(all_args[i + 1])
        if all_args[0] is None:
            raise RuntimeError("No experiment directory defined in constructor or .env!\n"
                               "Define in .env using MLM_EXPERIMENT_DIR=/PATH/TO/DIR")
        self.experiment_dir = os.path.join(all_args[0], 'mnemonic-experiments')
        self.hyper_experiment_dir = os.path.join(all_args[0], 'mnemonic-hyper-experiments')

        if not os.path.exists(self.experiment_dir):
            os.mkdir(self.experiment_dir)
        if not os.path.exists(self.hyper_experiment_dir):
            os.mkdir(self.hyper_experiment_dir)

        if all_args[2] is None:
            raise RuntimeError("No checkpoint directory defined in constructor or .env!\n"
                               "Define in .env using MLM_CHECKPOINT_DIR=/PATH/TO/CHECKPOINTS")
        self.checkpoint_dir = os.path.join(all_args[2], '.mnemonic-checkpoints')
        if not os.path.exists(self.checkpoint_dir):
            os.mkdir(self.checkpoint_dir)

        self.config_dir = all_args[4]
        if self.config_dir is None:
            warnings.warn("MLM_CONFIG_DIR not defined!")
        self.secret_root = all_args[6]
        if self.secret_root is None:
            warnings.warn("MLM_SECRET_ROOT_DIR not defined!")

    def clear_caches(self, prompt=True):
        """
        :param prompt:
        Prompt the user for decision

        Deletes everything in the checkpoints directory
        """
        while prompt:
            resp = input(f"This will delete everything in {self.checkpoint_dir}! Are you sure? [y/n] ")
            if resp == 'n':
                return
            elif resp == 'y':
                break
        for f in os.listdir(self.checkpoint_dir):
            real_dir = os.path.join(self.checkpoint_dir, f)
            if os.path.isfile(real_dir):
                os.remove(real_dir)
            else:
                shutil.rmtree(real_dir)

    def retrieve_experiment_runner(self, cache_token: str) -> ExperimentRunner:
        """
        Based on the cache token available at the checkpoints directory
        it retrieves a runner with all its cached data using cache_token
        """

        meta_cache = RunnerCache(directory=self.checkpoint_dir,
                            token=f'{cache_token}-META')

        meta_cache.LOAD()

        meta = meta_cache.SET_IFN('RUNNER_CONSTRUCTOR_META', None)
        cfg_base = meta_cache.SET_IFN_CFG('RUNNER_CFG_BASE_META', None)

        if meta is None or cfg_base is None:
            raise FileNotFoundError(f"No runner with cache token {cache_token} found in {self.checkpoint_dir}")

        runner = ExperimentRunner(experiment_dir=meta['experiment_dir'],
                                  checkpoint_dir=self.checkpoint_dir,
                                  verbose=meta['verbose'],
                                  cache_token=cache_token,
                                  cfg_path=meta['cfg_path'],
                                  cfg_base=cfg_base,
                                  secret_root=self.secret_root,
                                  meta_cache=meta_cache)

        preprocessing_pipeline = meta_cache.SET_IFN('RUNNER_PREP_PIPELINE_META', [])
        recurring_pipeline = meta_cache.SET_IFN('RUNNER_RECURRING_PIPELINE_META', [])
        run_func = meta_cache.SET_IFN('RUNNER_RUN_META', None)
        for func in preprocessing_pipeline:
            runner.preprocessing_pipeline.update_function(func)
        for func in recurring_pipeline:
            runner.recurring_pipeline.update_function(func)
        if run_func is not None:
            runner.implement_run(run_func)
        return runner

    def create_experiment_runner(self, verbose: int = 0,
                                 description: str = 'description not specified!',
                                 cfg_base: Optional[ConfigurationNode] = None,
                                 experiment_name: Optional[str] = None,
                                 cfg_dir: Optional[str] = None,
                                 cache_token: Optional[str] = None) -> ExperimentRunner:

        experiment_path = _get_new_experiment_path(self.experiment_dir, experiment_name)
        with open(os.path.join(experiment_path, 'DESCRIPTION.txt'), 'w') as f:
            f.write(description)

        cfg_path = None
        if cfg_dir is not None:
            if self.config_dir is not None:
                cfg_path = os.path.join(self.config_dir, cfg_dir)
            else:
                cfg_path = cfg_dir

        meta_cache = RunnerCache(directory=self.checkpoint_dir, token=f'{cache_token}-META')
        meta_cache.LOAD()

        ret = ExperimentRunner(experiment_dir=experiment_path,
                               checkpoint_dir=self.checkpoint_dir, verbose=verbose,
                               cache_token=cache_token,
                               cfg_path=cfg_path,
                               cfg_base=copy.deepcopy(cfg_base),
                               secret_root=self.secret_root,
                               meta_cache=meta_cache)
        meta = {
            'experiment_dir': ret.experiment_dir,
            'verbose': ret.verbose,
            'cfg_path': ret.cfg_path
        }

        meta_cache.SET('RUNNER_CONSTRUCTOR_META', meta)
        meta_cache.SET_CFG('RUNNER_CFG_BASE_META', ret.cfg)
        meta_cache.SAVE()
        return ret

    def retrieve_hyper_experiment_runner(self, cache_token: str) -> HyperExperimentRunner:
        cache = RunnerCache(directory=self.checkpoint_dir,
                            token=f'{cache_token}-HYPER-META')
        cache.LOAD()
        meta = cache.SET_IFN('HYPER_CONSTRUCTOR_META', None)
        cfg_base = cache.SET_IFN_CFG('CFG_BASE_HYPER_META', None)

        if meta is None or cfg_base is None:
            raise FileNotFoundError(f"Metadata for hyper runner not found in {self.checkpoint_dir}")

        runners = []
        for runner_tokens in meta['runner_cache_tokens']:
            runners.append(self.retrieve_experiment_runner(runner_tokens))

        return HyperExperimentRunner(
            cfg_palette_dir=meta['cfg_palette_dir'],
            cfg_base=cfg_base,
            hyper_experiment_path=meta['hyper_experiment_path'],
            cache_token=f'{cache_token}-HYPER',
            verbose=meta['verbose'],
            secret_root=self.secret_root,
            checkpoint_dir=self.checkpoint_dir,
            experiment_runners=runners
        )

    def create_hyper_experiment_runner(self,
                                       cfg_base: ConfigurationNode,
                                       cfg_palette_path: str,
                                       experiment_runners: List[ExperimentRunner],
                                       verbose: int = 0,
                                       description: str = 'description of hyper experiment not specified!',
                                       cache_token: Optional[str] = None,
                                       experiment_name: Optional[str] = None
                                       ) -> HyperExperimentRunner:
        experiment_name = 'hyper-exp' if experiment_name is None else experiment_name
        hyper_experiment_path = _get_new_experiment_path(self.hyper_experiment_dir, experiment_name)
        with open(os.path.join(hyper_experiment_path, 'DESCRIPTION.txt'), 'w') as f:
            f.write(description)
        cfg_palette_path = os.path.join(self.config_dir, cfg_palette_path)
        ret = HyperExperimentRunner(cfg_palette_dir=cfg_palette_path,
                                    cfg_base=cfg_base,
                                    hyper_experiment_path=hyper_experiment_path,
                                    cache_token=f'{cache_token}-HYPER',
                                    verbose=verbose,
                                    secret_root=self.secret_root,
                                    checkpoint_dir=self.checkpoint_dir,
                                    experiment_runners=experiment_runners)
        meta_data = {
            'hyper_experiment_path': hyper_experiment_path,
            'cfg_palette_dir': cfg_palette_path,
            'verbose': verbose,
            'runner_cache_tokens': [x.CACHE.TOKEN for x in experiment_runners]
        }
        meta_cache = RunnerCache(directory=ret.checkpoint_dir, token=f'{ret.CACHE.TOKEN}-META')
        meta_cache.SET('HYPER_CONSTRUCTOR_META', meta_data)
        meta_cache.SET_CFG('CFG_BASE_HYPER_META', cfg_base)
        meta_cache.SAVE()
        return ret


FACTORY = RunnerFactory()



import os
from typing import Dict


def get_all_codes(parent_directory: str) -> Dict[str, str]:
    """
    Given a parent_directory it outputs a dictionary mapping
    the codes to their corresponding yaml file
    """
    all_codes = {}
    for f in os.listdir(parent_directory):
        if f.split('.')[-1] == 'yaml':
            all_codes[f.split('-MLM-')[0]] = os.path.join(parent_directory, f)

    return all_codes



def _register_generic(module_dict, module_name, module):
    assert module_name not in module_dict
    module_dict[module_name] = module


# Inherits from dict
class Registry(dict):
    """
    A helper class for managing registering modules, it extends a dictionary
    and provides register functions.
    Access of module is just like using a dictionary, eg:
        f = some_registry["foo_module"]
    """

    # Instanciated objects will be empyty dictionaries
    def __init__(self, *args, **kwargs):
        super(Registry, self).__init__(*args, **kwargs)

    # Decorator factory. Here self is a Registry dict
    def register(self, module_name, module=None):
        # Inner function used as function call
        if module is not None:
            _register_generic(self, module_name, module)
            return

        # Inner function used as decorator -> takes a function as argument
        def register_fn(fn):
            _register_generic(self, module_name, fn)
            return fn

        return register_fn  # decorator factory returns a decorator function
